# TP Hachage



 
# Partie 1 - Réaliser une fonction de hachage minimale parfaite 

On travaillera sur `tp_2_miso_mphf.py`.

Tout d'abord, voyons ensemble l'idée de la construction de la MPHF.
Puis suivez les commentaires dans le code à compléter.
Vous devez 
- finaliser la fonction pour construire la MPHF (`construction_mphf`), 
- écrire la fonction pour obtenir le hash d'un élément par la MPHF (`get_hash_mphf`), 
- écrire la fonction pour créer une table de hachage avec cette mphf (`create_hash_table`)

Ecrivez vos réponses et commentaires dans ce document.

Puis décommenter `compare_taille` à la fin et expliquer les résultats.

Bonus : faites varier `nb_niveaux` et `gamma`, voyez quelle influence ils peuvent avoir.

Lorsqu'on augmente ces deux paramètres, la taille de la table avec MPHF n'évolue pas (ce qui est normal puisque la taille de la table ne dépend que de la taille de l'entrée).
Cependant, on remarque un allongement du temps d'exécution de la fonction lorsque les paramètres ont de grandes valeurs.

# Partie 2 - Analyse de performance de dictionnaires en Python

On travaillera sur `tp_2_miso_dict.py`

A la fin de cette partie, en lançant `python tp_2_miso_dict.py` on doit obtenir 4 graphiques.

## Introduction

Nous allons analyser la performance de dictionnaires `dict` en Python en fonction de différents facteurs : que le facteur de charge (loading factor), 
le temps d'insertion et la taille de la mémoire occupée. Vous pourrez utiliser matplotlib et numpy pour visualiser les résultats.

##  Étude du facteur de charge

Fonction `experiment_load_factor`:

0. La fonction `experiment_load_factor` doit être définie pour prendre en entrée une liste de facteurs de charge (`load_factors`) et renvoyer les temps 
- d'insertion de clefs, 
- les nombres de réallocations de mémoire (quand le dictionnaire est ré-écrit en mémoire pour agrandir sa taille) 
- et les tailles de mémoire occupées par le dictionnaire pour chaque facteur de charge. 

1. Initialisation des listes pour stocker les résultats.
Initialiser dans la fonction `experiment_load_factor` des listes `insertion_times`, `num_resizes` et `sizes` seront utilisées pour stocker les trois résultats décrits au dessus.

2. La fonction `experiment_load_factor` doit boucler sur chaque facteur de charge de la liste en entrée `load_factors`. Créer un dictionnaire d'abord vide à chaque fois.

3. Initialiser les variables `num_elements`, `start_time`, `num_resize` et `last_size` pour respectivement mesurer  le nombre d'éléments, le temps d'insertion et le nombre de réallocations de mémoire.

4. Insérer les éléments dans le dictionnaire et mesurer le temps d'insertion pour chaque élément, vérifier le nombre de réallocations mémoire (utiliser par exemple `time.time()` et `sys.getsizeof()` pour mesurer le temps et la taille du dictionnaire avant et après chaque insertion).

5. Pour un facteur de charge donné, stocker les résultats dans les listes `insertion_times`, `num_resizes` et `sizes`.

## Deuxième étude

6. A quoi sert la fonction `experiment_longest` ?

## Visualisation des résultats

7. Créez quatre graphiques au format png ou pdf : 

- Un graphique du temps d'insertion en fonction du facteur de charge (obtenu question 5)
- Un graphique du nombre de réallocations de mémoire en fonction du facteur de charge (obtenu question 5)
- Un graphique de la taille de mémoire occupée en fonction du nombre d'éléments (obtenu question 5)
- Un histogramme des fréquences des temps d'insertions discrétisés (code fourni, remplacer la liste vide par la bonne entrée)

10. Commentez vos résultats.

Le temps d'insertion de n éléments est proportionnel à n, comme le montre le Plot1. Cela signifie que l'insertion d'un élément dans un dictionnaire prend quasiment toujours le même temps, peu importe le nombre d'élements déjà présents dans un dictionnaire.
Ceci est dû à la manière dont la mémoire est allouée lors de la création du dictionnaire. En effet, lors de la création d'un dictionnaire, 8 cellules de mémoire vides et contigües sont allouées, ce qui permet un enregistrement des informations uniforme tant que suffisamment d'espace reste vide, l'adresse de stockage à l'intérieur de cette zone de mémoire étant déduite du hachage de la clé. Cependant, lorsqu'un seuil de remplissage de cet espace alloué est atteint, l'ensemble des informations contenues dans cette zone de mémoire sont dupliquées dans un espace deux fois plus grand, ce qui permet de continuer d'agrandir le dictionnaire. Cette opération est coûteuse en temps mais est de plus en plus rare au fur et à mesure de l'agrandissement du dictionnaire, ce qui permet de conserver une complexité en temps linéaire.

Le nombre de réallocations de mémoire augmente de moins en moins vite au fur et à mesure de l'agrandissement de la taille de l'entrée, comme le montre le Plot2. En effet, comme à chaque agrandissement d'espace alloué celui-ci est doublé, à chaque agrandissement de plus en plus d'éléments peuvent être enregistrés, rendant le besoin de resizing moins fréquent.

Ceci a pour conséquence une taille de mémoire allouée qui évolue "en escaliers", comme le montre le Plot3. En effet, chaque doublement d'espacé alloué se traduit par une "marche" deux fois plus haute sur le graphique, mais le besoin de resizing étant de moins en moins fréquent au fur et à mesure de l'agrandissement du dictionnaire, ces "marches" sont de plus en plus longues.

Enfin, l'histogramme montre que le temps d'insertion est quasiment toujours identique pour toutes des valeurs du dictionnaire, comme ce qu'on avait déduit du Plot1, excepté dans de rares cas où le temps d'insertion est plus long, ce qui peut correspondre à un moment où l'insertion de valeur est précédée d'un doublement d'allocation d'espace mémoire.