import matplotlib.pyplot as plt
import numpy as np
import time
import sys


###### PARTIE 2 ######

def experiment_load_factor(load_factors: list[float]) -> tuple[list[float], list[int], list[int]]:
	"""
	Étude du facteur de charge
	:param load_factors: Liste de facteurs de charges (représentent nombre d'éléments dans dict).
	:return: Trois listes représentant pour chaque facteur de charge: le temps de création du dict, le nombre de changements de taille, et l'espace occupé par le dict
	"""
	insertion_times = []
	resizing = []
	memory_sizes = []
	for i in load_factors:
		d = {}
		resizing_count = 0
		last_memory_size = sys.getsizeof(d)
		num_elements = int(i*100)
		start_time = time.perf_counter()
		for j in range(num_elements):
			d[j] = j
			current_memory_size = sys.getsizeof(d)
			if current_memory_size != last_memory_size :
				resizing_count += 1
				last_memory_size = current_memory_size
		end_time = time.perf_counter()
		insertion_times.append(end_time - start_time)
		resizing.append(resizing_count)
		memory_sizes.append(last_memory_size)
	return insertion_times, resizing, memory_sizes


def experiment_longest():
	"""
	Mesure 10000 temps d'insertion de couple clé + valeur dans le dictionnaire et groupe les temps d'insertion dans une liste de fréquences.
	:return: Liste de groupes de temps d'insertion avec, pour chaque groupe, son effectif.
	"""
	d = {}
	insertion_times = []

	for i in range(10000):
		key = str(i)
		value = i
		start_time = time.time()
		d[key] = value
		insertion_time = time.time() - start_time
		insertion_times.append(insertion_time)
	frequencies = np.histogram(insertion_times)[0]
	return frequencies

def visualisation(load_factors, insertion_times, num_resizes, sizes, frequencies):
	"""
	Visualisation des résultats
	"""
	# Temps d'insertion en fonction du facteur de charge
	plt.scatter(load_factors, insertion_times)
	plt.xlabel('Facteurs de charge')
	plt.ylabel('Temps d\'insertion (s)')
	plt.title('Temps d\'insertion en fonction du facteur de charge')
	plt.savefig('Plot1.png')
	# Nombre de réallocations de mémoire en fonction du facteur de charge
	plt.clf()
	plt.scatter(load_factors, num_resizes)
	plt.xlabel('Facteurs de charge')
	plt.ylabel('Nombre de resizings')
	plt.title('Nombre de réallocations de mémoire en fonction du facteur de charge')
	plt.savefig('Plot2.png')
	# Taille de mémoire occupée en fonction du nombre d'éléments
	plt.clf()
	plt.scatter([i * 100 for i in load_factors], sizes)
	plt.xlabel('Nombre d\'élements')
	plt.ylabel('Taille de mémoire occupée')
	plt.title('Taille de mémoire occupée en fonction du nombre d\'éléments')
	plt.savefig('Plot3.png')
	# Deuxième étude
	f = frequencies
	plt.figure(figsize=(10, 6))
	plt.bar(range(len(f)), f)
	plt.xlabel('Temps d\'insertion (s)')
	plt.ylabel('Fréquence')
	plt.title('Histogramme des fréquences des temps d\'insertions')
	plt.yscale('log')
	xticks = np.logspace(-6, 1, 3)  
	xtick_labels = [f'{x:.1e}' for x in xticks]  
	plt.xticks(xticks, xtick_labels)    
	plt.savefig('histogramme.png')

load_factors = [0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
insertion_times, num_resizes, sizes = experiment_load_factor(load_factors)
frequencies = experiment_longest()
visualisation(load_factors, insertion_times, num_resizes, sizes, frequencies)
